use mongod::{
    assert_reference_of,
    derive::{Model, Referencable},
    Model, Referencable, Reference, Validate,
};
use mongodb::bson::doc;
use serde::{Deserialize, Serialize};
use serde_json::json;

use crate::{library::artist::Artist, route::ToAPI};

use super::track::Track;

#[derive(Debug, Clone, Serialize, Deserialize, Model, Referencable)]
pub struct Album {
    pub _id: String,
    pub title: String,
    pub artist_id: Option<Reference>,
}

impl Album {
    pub async fn create(title: &str, artist: Option<&str>) -> Self {
        let a = Self {
            _id: uuid::Uuid::new_v4().to_string(),
            title: title.to_string(),
            artist_id: if let Some(artist) = artist {
                Some(Reference::new_raw(artist).await.unwrap())
            } else {
                None
            },
        };
        a.insert().await.unwrap();
        a
    }

    pub async fn get_tracks_of_album(album: &str) -> Vec<Track> {
        Track::find(doc! { "album_id": album}, None).await.unwrap()
    }

    /// Returns the cover image of an album, or `None` if it doesn't exist.
    ///
    /// This method first retrieves the list of tracks in the album, then looks for a file named "cover.{ext}" where {ext} is one of png, jpg, jpeg, avif. The first existing cover file found will be returned.
    pub async fn get_cover(&self) -> Option<String> {
        let track_path = Self::get_tracks_of_album(&format!("album::{}", self._id))
            .await
            .first()?
            .path
            .clone();
        let track_path = std::path::Path::new(&track_path);

        for ext in ["png", "jpg", "jpeg", "avif", "webp"] {
            let cover_file = track_path.parent()?.join(format!("cover.{ext}"));

            if cover_file.exists() {
                return Some(cover_file.to_str().unwrap().to_string());
            }
        }

        None
    }
}

impl ToAPI for Album {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": &self._id,
            "title": &self.title,
            "artist": self.artist_id.as_ref().map(mongod::Reference::id),
            "cover_url": if self.get_cover().await.is_some() {
                Some(format!("/album/{}/cover", self._id))
            } else {
                None
            }
        })
    }
}

impl Validate for Album {
    async fn validate(&self) -> Result<(), String> {
        if let Some(artist_id) = &self.artist_id {
            assert_reference_of!(artist_id, Artist);
        }

        Ok(())
    }
}
