use mongod::{
    derive::{Model, Referencable},
    reference_of, Model, Referencable, Validate,
};
use mongodb::bson::doc;
use serde::{Deserialize, Serialize};
use serde_json::json;

use crate::route::ToAPI;

use super::track::Track;

#[derive(Debug, Clone, Serialize, Deserialize, Model, Referencable)]
pub struct Artist {
    pub _id: String,
    pub name: String,
}

impl Artist {
    pub async fn create(name: &str) -> Self {
        let a = Artist {
            _id: uuid::Uuid::new_v4().to_string(),
            name: name.to_string(),
        };
        a.insert().await.unwrap();
        a
    }

    /// Gets the image of an artist or `None` if it can't be found.
    ///
    /// This function gets a track from the artist. It then expects the folder structure to be `Artist/Album/Track.ext` and searches for an image file named `artist` in the artist folder.
    pub async fn get_image_of(id: &str) -> Option<String> {
        let track_path = Track::find_one(doc! { "artist_id": reference_of!(Artist, id)})
            .await?
            .path;
        let track_path = std::path::Path::new(&track_path);

        let artist_path = track_path.parent()?.parent()?;

        for ext in ["png", "jpg", "jpeg", "avif", "webp"] {
            let cover_file = artist_path.join(format!("artist.{ext}"));

            if cover_file.exists() {
                return Some(cover_file.to_str().unwrap().to_string());
            }
        }

        None
    }
}

impl ToAPI for Artist {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": &self._id,
            "name": &self.name,
            "image": if Artist::get_image_of(self.id()).await.is_some() {
                Some(format!("/artist/{}/image", self.id()))
            } else {
                None
            }
        })
    }
}

impl Validate for Artist {
    async fn validate(&self) -> Result<(), String> {
        Ok(())
    }
}
